@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# -*- coding: utf-8 -*-
 #
 #  date_frames.py
 """Mathematical Operations for Data Frames"""
@@ -7,8 +6,8 @@
 #  Copyright 2019 Dominic Davis-Foster <dominic@davis-foster.co.uk>
 #
 #  Based on
-#		http://jonathansoma.com/lede/foundations/classes/pandas%20columns%20and%20functions/apply-a-function-to-every-row-in-a-pandas-dataframe/
-#		Copyright 2016 Jonathan Soma
+# 		http://jonathansoma.com/lede/foundations/classes/pandas%20columns%20and%20functions/apply-a-function-to-every-row-in-a-pandas-dataframe/
+# 		Copyright 2016 Jonathan Soma
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -25,14 +24,25 @@
 #  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 #  MA 02110-1301, USA.
 #
-#
+
+# stdlib
+from math import log, log10
+from typing import List, Optional, Sequence
+
+# 3rd party
+import numpy  # type: ignore
+from pandas import Series  # type: ignore
+
+# this package
+from . import outliers
 
 # Outlier Modes
 MAD = 1
 QUARTILES = 2
 STDEV2 = 3
 
-def df_mean(row, column_label_list=None):
+
+def df_mean(row: Series, column_label_list: Optional[Sequence[str]] = None) -> float:
 	"""
 	Calculate the mean of each row for the specified columns of a data frame
 
@@ -42,75 +52,63 @@ def df_mean(row, column_label_list=None):
 				["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to calculate mean for
-	:type column_label_list: list
 
 	:return: Mean
 	:rtype: float
 	"""
-	
-	from numpy import nanmean
-	
+
 	if column_label_list is None:
 		column_label_list = list(row.index)
-	
-	return nanmean(row[column_label_list])
+
+	return float(numpy.nanmean(row[column_label_list]))
 
 
-def df_median(row, column_label_list=None):
+def df_median(row: Series, column_label_list: Optional[Sequence[str]] = None) -> float:
 	"""
 	Calculate the median of each row for the specified columns of a data frame
 
 	Do not call this function directly; use it with df.apply() instead.
 
 	data_frame["Median"] = data_frame.apply(df_median, args=(
-				["Bob", "Alice"],), axis=1)
+							["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to calculate median for
-	:type column_label_list: list
 
 	:return: Median
 	:rtype: float
 	"""
-	
-	from numpy import nanmedian
-	
+
 	if column_label_list is None:
 		column_label_list = list(row.index)
-	
-	return nanmedian(row[column_label_list])
+
+	return float(numpy.nanmedian(row[column_label_list]))
 
 
-def df_stdev(row, column_label_list=None):
+def df_stdev(row: Series, column_label_list: Optional[Sequence[str]] = None) -> float:
 	"""
 	Calculate the standard deviation of each row for the specified columns of a data frame
 
 	Do not call this function directly; use it with df.apply() instead.
 
 	data_frame["Stdev"] = data_frame.apply(df_stdev, args=(
-				["Bob", "Alice"],), axis=1)
+							["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to calculate standard deviation for
-	:type column_label_list: list
 
 	:return: Standard deviation
 	:rtype: float
 	"""
-	
-	from numpy import nanstd
-	
+
 	if column_label_list is None:
 		column_label_list = list(row.index)
-	
-	return nanstd(row[column_label_list])
+
+	return float(numpy.nanstd(row[column_label_list]))
 
 
-def df_log_stdev(row, column_label_list=None):
+def df_log_stdev(row: Series, column_label_list: Optional[Sequence[str]] = None) -> float:
 	"""
 	Calculate the standard deviation of the log10 values in each row for the specified columns of a data frame
 
@@ -120,24 +118,19 @@ def df_log_stdev(row, column_label_list=None):
 				["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to calculate standard deviation for
-	:type column_label_list: list
 
 	:return: Standard deviation
 	:rtype: float
 	"""
-	
-	from numpy import nanstd, nan
-	from math import log10
-	
+
 	if column_label_list is None:
 		column_label_list = list(row.index)
-	
-	return nanstd([log10(x) if x > 0.0 else nan for x in row[column_label_list]])
+
+	return float(numpy.nanstd([log10(x) if x > 0.0 else numpy.nan for x in row[column_label_list]]))
 
 
-def df_percentage(row, column_label, total):
+def df_percentage(row: Series, column_label: str, total: float) -> float:
 	"""
 	Returns the value of the specified column as a percentage of the given total
 	The total is usually the sum of the specified column
@@ -148,20 +141,19 @@ def df_percentage(row, column_label, total):
 				13, "Bob"), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label: column label to calculate percentage for
 	:type column_label: str
 	:param total: total value
-	:type column_label: str
+	:type total: float
 
 	:return: Percentage * 100
 	:rtype: float
 	"""
-	
+
 	return (row[column_label] / float(total)) * 100.0
 
 
-def df_log(row, column_label_list, base=10):
+def df_log(row: Series, column_label_list: Sequence[str], base: float = 10) -> float:
 	"""
 	Calculate the logarithm of the values in each row for the specified columns of a data frame
 
@@ -171,9 +163,7 @@ def df_log(row, column_label_list, base=10):
 				["Bob"],10), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to calculate log for
-	:type column_label_list: list
 	:param base: logarithmic base
 	:type base: float
 
@@ -181,15 +171,13 @@ def df_log(row, column_label_list, base=10):
 	:rtype: float
 	"""
 
-	from math import log
-	
 	if all(row[column_label_list][i] > 0.0 for i in range(len(row[column_label_list]))):
 		return log(row[column_label_list], base)
 	else:
 		return 0
 
 
-def df_data_points(row, column_label_list):
+def df_data_points(row: Series, column_label_list: Sequence[str]) -> List:
 	"""
 	Compile the values for the specified columns in each row into a list
 
@@ -199,18 +187,16 @@ def df_data_points(row, column_label_list):
 				["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to calculate standard deviation for
-	:type column_label_list: list
 
 	:return: data points
 	:rtype: list
 	"""
-	
+
 	return [row[column_label] for column_label in column_label_list]
 
 
-def df_outliers(row, column_label_list=None, outlier_mode=MAD):
+def df_outliers(row: Series, column_label_list: Sequence[str] = None, outlier_mode: int = MAD) -> Series:
 	"""
 	Identify outliers in each row
 
@@ -220,39 +206,33 @@ def df_outliers(row, column_label_list=None, outlier_mode=MAD):
 				["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to determine outliers for
-	:type column_label_list: list
 	:param outlier_mode: outlier detection method to use
 	:type outlier_mode: int
 
 	:return: outliers
-	:rtype: pandas.core.series.Series
 	"""
-	
-	import pandas as pd
-	from . import outliers
-	
+
 	if column_label_list is None:
 		column_label_list = list(row.index)
-	
+
 	data = row[column_label_list]
 	if all(all(y == 0.0 for y in x) for x in data):
-		return pd.Series([[], [0.0]*len(data[0])])
-	
+		return Series([[], [0.0] * len(data[0])])
+
 	if outlier_mode == MAD:
 		x = outliers.mad_outliers(data)
 	elif outlier_mode == QUARTILES:
 		x = outliers.quartile_outliers(data)
 	elif outlier_mode == STDEV2:
-		x = outliers.stdev_outlier(data, 2)  # outlier classed as more than 2 stdev away from mean
+		x = outliers.stdev_outlier(data, rng=2)  # outlier classed as more than 2 stdev away from mean
 	else:
-		return None
-	
-	return pd.Series(list(x))
+		raise ValueError("Unknown outlier mode.")
+
+	return Series(list(x))
 
 
-def df_count(row, column_label_list=None):
+def df_count(row: Series, column_label_list: Optional[Sequence[str]] = None) -> int:
 	"""
 	Count the number of occurrences of a non-NaN value in the specified columns of a data frame
 
@@ -262,22 +242,17 @@ def df_count(row, column_label_list=None):
 				["Bob", "Alice"],), axis=1)
 
 	:param row: row of the data frame
-	:type row: pandas.core.series.Series
 	:param column_label_list: list of column labels to count occurrences in
-	:type column_label_list: list
 
 	:return: Count of the occurrences of non-NaN values
 	:rtype: int
 	"""
-	
-	import numpy
-	
+
 	if column_label_list is None:
 		column_label_list = list(row.index)
-	
+
 	count = 0
 	for column_label in column_label_list:
 		if row[column_label] and not numpy.isnan(row[column_label]):
 			count += 1
 	return count
-
